/*
  This file contains various utility functions for MAPSIM v0.1

 $Author: $
 $Date: $
 $Id:$
 $Log:$

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "nrutil.h"
#include "mapsim.h"

/* generate legengre polynomials (times a normalization) */
void M_gen(double x,double **M,int lmax)
{
  int m,l;
  double sq,fac,crap1,crap2,crap3;
  double sqrtl[2*LMAX+4];

  /* Initialize M */
  for(l=0 ; l<=lmax ; l++)
    for(m=0 ; m<=lmax ; m++)
      M[l][m]=0;
  
  for(l=0 ; l<=2*lmax+3 ; l++)
    sqrtl[l] = sqrt((double)l);
  sq=sqrt((double)(1-x*x));
  M[0][0]=pow(4*PI,-.5);

  for(m=0 ; m<=lmax ; m++){
  
    /* l=m */
    if(m>0)
      M[m][m]=sqrt((double)(2*m+1))*M[m-1][m-1]*sq/sqrt((double)(2*m));
    
    /* l=m+1 */
    if(m<lmax)
      M[m+1][m]=x*sqrtl[2*m+3]*M[m][m];
    
    /* l>=m+2 */
    for(l=m+2 ; l<=lmax ; l++)
      M[l][m]=(x*(2*l-1)*(sqrtl[2*l+1]*sqrtl[l-m]/
			  (sqrtl[2*l-1]*sqrtl[l+m]))*M[l-1][m]-
	       (l+m-1)*(sqrtl[2*l+1]*sqrtl[l-m]*sqrtl[l-m-1]/
			(sqrtl[2*l-3]*sqrtl[l+m]*sqrtl[l+m-1]))*
	       M[l-2][m])/((float)l-m);}
  
  /* Fix up minus sins */
  for(l=1 ; l<=lmax ; l++)
    for(m=1 ; m<=l ; m+=2) M[l][m]*=-1.0;
}

/* Generate real valued spherical harmonics */
void Y_gen(double **M,double **Y_p,double **Y_n,
	   double *cosine,double *sine,int lmax)
{
  
  int l,m;
  int i;
  double sq2;

  for(l=0 ; l<=lmax ; l++){
    Y_p[l][0]=M[l][0]*cosine[0];
    Y_n[l][0]=M[l][0]*sine[0];
    for(m=1 ; m<=l ; m++){
      Y_p[l][m]=M[l][m]*cosine[m];
      Y_n[l][m]=M[l][m]*sine[m];}}
}

/* generate sin[m*phi] and cos[m*phi] arrays */
void trig(double x,double y,double cosine[],double sine[],int lmax)
{
  int i;
  double d,sq;

  d=sqrt(x*x+y*y);
  sq=sqrt(2.0);
  
  cosine[0]=1;
  cosine[1]=x/d;
  sine[0]=0;
  sine[1]=y/d;
  
  for(i=2 ; i<lmax ; i++){
    cosine[i]=cosine[1]*cosine[i-1]-sine[1]*sine[i-1];
    sine[i]=sine[1]*cosine[i-1]+cosine[1]*sine[i-1];}
  return;
}

/* convert celestial coordinates to Euclidean x-y-z */
void rdtoxyz(double ra[],double dec[],double x[],double y[],double z[],
	     int npix)
{
  int i;
  double theta;

  for(i=0 ; i<npix ; i++){
    theta=90-dec[i]; 
    x[i]=sin(theta*PI/180)*cos(ra[i]*PI/180);
    y[i]=sin(theta*PI/180)*sin(ra[i]*PI/180);
    z[i]=cos(theta*PI/180);}
  return;
}

/* generate real valued W_lm, see equation 2.24, KKS.  */
void W_gen(double x,double **M,double **W,int lmax)
{
  double sint,cost;
  double sqe;
  double *Nl;
  double fac1,fac2,sfac;
  double ld,md;
  int l,m;

  Nl=dvector(0,lmax);

  cost=x;
  sint=sqrt((double)(1-x*x));
  sqe=sqrt((double)8);

  /* get sqrt[(l-2)!/(l+2)!] */
  Nl[0]=Nl[1]=0;
  for(l=2 ; l<=lmax ; l++){
    ld=(double)l;
    Nl[l]=sqrt((double)(1/((ld-1)*ld*(ld+1)*(ld+2))));}
  
  W[0][0]=W[1][0]=W[0][1]=W[1][1]=0;
  for(l=2 ; l<=lmax ; l++)
    for(m=0 ; m<=l ; m++){
      W[l][m]=0;
      
      ld=(double)l;
      md=(double)m;

      fac1=((ld-md*md)/(sint*sint)+(ld*(ld-1))/2.0);
      fac2=(ld+md)*(cost/(sint*sint));
      sfac=sqrt((double)(((2*ld+1)*(ld-md))/((2*ld-1)*(ld+md))));
      
      W[l][m]=-1.0*fac1*sqe*Nl[l]*M[l][m]+fac2*sqe*Nl[l]*sfac*M[l-1][m];
      W[l][m]*=-1.0;
    }
  return;
}

/* generate real valued X_lm, see equation 2.25, KKS.  */
void X_gen(double x,double **M,double **X,int lmax)
{
  double sint,cost;
  double sqe;
  double *Nl;
  double fac1,fac2,sfac;
  double ld,md;
  int l,m;
  
  Nl=dvector(0,lmax);
  
  cost = x;
  sint = sqrt((double)(1-x*x));
  sqe  = sqrt((double)8);
  
  /* get sqrt[(l-2)!/(l+2)!] */
  Nl[0]=Nl[1]=0;
  for(l=2 ; l<=lmax ; l++){
    ld=(double)l;
    Nl[l]=sqrt(1/((ld-1)*ld*(ld+1)*(ld+2)));}
  
  X[0][0]=X[1][0]=X[0][1]=X[1][1]=0;
  for(l=2 ; l<=lmax ; l++)
    for(m=0 ; m<=l ; m++){
      X[l][m]=0;
      
      ld=(double)l;
      md=(double)m;
      
      fac1=md*(ld-1)*cost/(sint*sint);
      fac2=md*(ld+md)/(sint*sint);
      sfac=sqrt(((2*ld+1)*(ld-md))/((2*ld-1)*(ld+md)));
      
      X[l][m]=-1*fac1*sqe*Nl[l]*M[l][m]+fac2*sqe*Nl[l]*sfac*M[l-1][m];
    }
  return;
}


/* return the determinant of a 3*3 matrix */
double det(double **M)
{
  double det=0.0;

  det =M[1][1]*(M[2][2]*M[3][3]-M[2][3]*M[3][2]);
  det+=M[1][2]*(M[2][3]*M[3][1]-M[2][1]*M[3][3]);
  det+=M[1][3]*(M[2][1]*M[3][2]-M[2][2]*M[3][1]);
  return det;
}
