
#include <iostream.h>
#include <fstream.h>
#include <math.h>
#include <vector.h>
#include "nrutil.h"
#include "mapsim.h"

#define ME 1e-30

main(int argc,char **argv)
{
  ofstream fout;
  
  double **W,**X,*p,**C;
  double **M,**Y_p,**Y_n;
  double **at_p,**at_n;
  double **ae_p,**ae_n;
  double **ab_p,**ab_n;
  double sqct,sqce,sqcb,l11,l21,l22;
  double *cosine,*sine;  
  double mt=0,mq=0,mu=0,di;
  double x1,x2,x3,sqe;
  double pq,nq,pu,nu;
  double t,q,u;

  int l,m,ni,my_n,my_s;  
  int i,j;
  int npix,lmax;
  int no_pe,my_pe;
  int size;

  long idum(-20);

  //
  // Welcome message
  // 
  system("clear");
  cout << endl;        
  cout << "************* " << endl;
  cout << " MAPSIM v1.0  " << endl;
  cout << "************* " << endl<<endl;
 
  //
  // Read in pixels from user 
  //
  Pixels pix;
  while(!pix.read_pixels());
  pix.setxyz();
  npix = pix.getNpix();
  
  //
  // Read Models from user
  //
  Models cl;
  while(!cl.read_models());
  lmax = cl.getlmax();

  //
  // Delare Maps
  //
  Maps map(npix);
  
  sine   = dvector(0,lmax);
  cosine = dvector(0,lmax);
   
  p = dvector(1,3);
  M = dmatrix(0,lmax,0,lmax);
  W = dmatrix(0,lmax,0,lmax);
  X = dmatrix(0,lmax,0,lmax);
   
  at_p=dmatrix(0,lmax,0,lmax);
  at_n=dmatrix(0,lmax,0,lmax);
  
  ae_p=dmatrix(0,lmax,0,lmax);
  ae_n=dmatrix(0,lmax,0,lmax);
  ab_p=dmatrix(0,lmax,0,lmax);
  ab_n=dmatrix(0,lmax,0,lmax);
  C=dmatrix(1,3,1,3);
  
  /***********************************/
  /* Generate a_lm's for T, E, and B */
  /***********************************/
  for(l=2 ; l<=lmax ; l++){

    C[1][1]         = cl.getTT(l);
    C[2][2]         = cl.getEE(l);
    C[3][3]         = cl.getBB(l);
    C[1][2]=C[2][1] = cl.getTE(l);
    C[1][3]=C[3][1] = cl.getTB(l);
    C[2][3]=C[3][2] = cl.getEB(l);
        
    if(det(C)<=0)
      if(my_pe==0){
	printf("ERROR: non-positive definate matrix\n");
	exit(-1);
      }

    choldc(C,3,p);
    C[1][1]=p[1];C[2][2]=p[2];C[3][3]=p[3];
    
    /* loop over m */
    for(m=0 ; m<=l ; m++){
      
      /* do a^t_lm and a^e_lm for m>=0 */
      x1=gasdev(&idum);
      x2=gasdev(&idum);
      x3=gasdev(&idum);
      at_p[l][m]=C[1][1]*x1;
      ae_p[l][m]=C[2][1]*x1+C[2][2]*x2;
      ab_p[l][m]=C[3][1]*x1+C[3][2]*x2+C[3][3]*x3;

      /* do a^t_lm and a^e_lm for m<0 */
      x1=gasdev(&idum);
      x2=gasdev(&idum);
      x3=gasdev(&idum);
      at_n[l][m]=C[1][1]*x1;
      ae_n[l][m]=C[2][1]*x1+C[2][2]*x2;
      ab_n[l][m]=C[3][1]*x1+C[3][2]*x2+C[3][3]*x3;
      
    }}
    
  /********************/
  /* loop over pixels */
  /********************/
  cout << " generating maps ..."<<endl;
  sqe = sqrt(2.0);
  for(i=0 ; i<npix ; i++){
    
    t = 0.0;
    q = 0.0;
    u = 0.0;
    
    /* calculate relevant quantities */
    trig(pix.getx(i),pix.gety(i),cosine,sine,lmax);
    M_gen(pix.getz(i),M,lmax);
    W_gen(pix.getz(i),M,W,lmax);
    X_gen(pix.getz(i),M,X,lmax);
    
    /* get T */
    for(l=2 ; l<=lmax ; l++){
      t += M[l][0] * at_p[l][0];
      for(m=1 ; m<=l ; m++){
	t += sqe * M[l][m] * cosine[m] * at_p[l][m];
	t -= sqe * M[l][m] * sine[m]   * at_n[l][m];}}
    
    /* get Q */
    for(l=2 ; l<=lmax ; l++){
      q -= ae_p[l][0] * W[l][0];
      for(m=0 ; m<=l ; m++){
	q -= W[l][m] * cosine[m] * ae_p[l][m];
	q += W[l][m] * sine[m]   * ae_n[l][m];
	q += X[l][m] * sine[m]   * ab_p[l][m];
	q += X[l][m] * cosine[m] * ab_n[l][m];
      }}
    
    /* get U */
    for(l=2 ; l<=lmax ; l++){
      u -= ab_p[l][0] * W[l][0];
      for(m=0 ; m<=l ; m++){
	u -= W[l][m] * cosine[m] * ab_p[l][m];
	u += W[l][m] * sine[m]   * ab_n[l][m];
	u -= X[l][m] * sine[m]   * ae_p[l][m];
	u -= X[l][m] * cosine[m] * ae_n[l][m];
      }}
        
    map.setT(t);
    map.setQ(q);
    map.setU(u);
    
    
#if defined(PRINT)
    printf("  T[%i] = %lf\tQ[%i] = %lf\tU[%i] = %lf\n",
	   i,map.getT(i),i,map.getQ(i),i,map.getU(i)); 
#endif
  
  }
  
  /**********************/
  /* write maps to file */
  /**********************/  
  map.setNpix();
  cout << "Generated "<<map.getNpix()
       <<" pixels in the maps ... writting to file ..."<<endl;
  fout.open("map.dat",ios::out);
  for(i=0 ; i<map.getNpix() ; i++)
    cout<<map.getT(i)<<map.getQ(i)<<map.getU(i)<<endl;
    
  return 0;
}







